/*
 * @(#)ExiValidator.java
 *
 * Copyright (C) 2011 The MITRE Corporation
 *
 * This program and its interfaces are free software;
 * you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

import javax.xml.XMLConstants;
import javax.xml.transform.sax.SAXSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;

import com.siemens.ct.exi.CodingMode;
import com.siemens.ct.exi.EXIFactory;
import com.siemens.ct.exi.FidelityOptions;
import com.siemens.ct.exi.GrammarFactory;
import com.siemens.ct.exi.api.sax.EXISource;
import com.siemens.ct.exi.exceptions.EXIException;
import com.siemens.ct.exi.exceptions.UnsupportedOption;
import com.siemens.ct.exi.grammar.Grammar;
import com.siemens.ct.exi.helpers.DefaultEXIFactory;

/** 
 * The ExiValidator class defines static methods (including main())
 * to demonstrate validation of EXI-encoded XML Instance Documents 
 * against a XML Schema. This class makes use of the built-in Java 
 * API in javax.xml, as well as the open-source EXIficient 
 * library (http://exificient.sourceforge.net/). This class also
 * contains code to validate typical, text-encoded XML to further 
 * illustrate how to integrate an EXI library into normal XML 
 * processing.
 * 
 * For more information on Efficient XML Interchange (XML), 
 * see http://www.w3.org/TR/2011/REC-exi-20110310/
 *
 * @author Craig Garrett, MITRE, garrett@mitre.org
 */
public class ExiValidator {

	public static void main(String[] args) {

		Map<ExiOption, String> commandLineArgs = null;
		if (args.length < 2)
			printHelp(); // this calls System.exit(0)

		commandLineArgs = parseCommandLineArgs(args); // this calls printHelp() and System.exit(0) if there is an error

		if (commandLineArgs.containsKey(ExiOption.VALIDATION_SCHEMA) && (commandLineArgs.containsKey(ExiOption.XML_IN) ^ commandLineArgs.containsKey(ExiOption.EXI_IN))) {
			if (commandLineArgs.containsKey(ExiOption.XML_IN)) {
				validate(commandLineArgs.get(ExiOption.XML_IN), commandLineArgs.get(ExiOption.VALIDATION_SCHEMA), null); // this calls System.exit(1) if validation fails
			}
			else if (commandLineArgs.containsKey(ExiOption.EXI_IN)) {
				EXIFactory exiFactory = createEXIFactory(commandLineArgs);
				validate(commandLineArgs.get(ExiOption.EXI_IN), commandLineArgs.get(ExiOption.VALIDATION_SCHEMA), exiFactory); // this calls System.exit(1) if validation fails
			}
			else
				printHelp(); // this calls System.exit(0)
		}
		else {
			printHelp(); // this calls System.exit(0)
		}
	}

	/**
	 * Validate an EXI-encoded or text-encoded XML Instance Document against a XML Schema.
	 * This method calls System.exit(1) if validation fails. If the exiFactory parameter 
	 * is null, then this method will interpret the XML Instance Doc as a regular, 
	 * text-encoded XML file.
	 * 
	 * @param instanceDocFilename the name of the file to be validated
	 * @param schemaFilename the name of the XML Schema
	 * @param exiFactory the object provided by EXIficient to generate XML Sources and ultimately XML Readers. If this is null, then this method will interpret the XML Instance Doc as a regular, text-encoded XML file.
	 */
	public static void validate(String instanceDocFilename, String schemaFilename, EXIFactory exiFactory) {
		File documentFile = new File(instanceDocFilename);
		File schemaFile = new File(schemaFilename);

		SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
		try {

			// create a SAXSource for the EXI Instance Document to be validated
			System.out.println("\tReading instance document: " + instanceDocFilename);
			InputStream inputStream = new FileInputStream(documentFile); // use a FileInputStream; using a FileReader results in EXIficient throwing "java.io.IOException: Stream closed" during the call to validator.validate(saxSource)
			org.xml.sax.InputSource saxInputSource = new InputSource(inputStream);
			javax.xml.transform.sax.SAXSource saxSource = new SAXSource(saxInputSource);

			if (exiFactory != null) {
				// Obtain an EXI Reader from EXIficient which implements the org.xml.sax.XMLReader interface
				EXISource exiSource = new EXISource(exiFactory); // get a EXISource from com.siemens.ct.exi.api.sax
				XMLReader exiReader = exiSource.getXMLReader(); // get an object provided by EXIficient which implements the org.xml.sax.XMLReader interface
				saxSource.setXMLReader(exiReader); // tell the javax.xml.transform.sax.SAXSource to use the XMLReader provided by EXIficient
			}
			
			// create a validating schema
			System.out.println("\tReading XML Schema: " + schemaFilename);
			Schema schema = schemaFactory.newSchema(schemaFile);
			Validator validator = schema.newValidator();

			// attempt to validate the XML Instance Document against the XML Schema
			System.out.println("\tValidating " + instanceDocFilename + " against " + schemaFilename + "\n");
			validator.validate(saxSource);

		} catch (SAXException e) {
			schemaValidationFail(e, instanceDocFilename); System.exit(1);
		} catch (FileNotFoundException e) {
			System.err.println(e.getMessage()); System.exit(1);
		} catch (IOException e) {
			System.err.println(e.getMessage()); System.exit(1);
		} catch (EXIException e) {
			System.err.println(e.getMessage()); System.exit(1);
		}
		System.out.println("\tResult: " + instanceDocFilename + " is schema-valid.");
	}

	/**
	 * Handle a SAXException, typically thrown when a 
	 * XML Instance Document fails validation against a 
	 * XML Schema.
	 * 
	 * @param e the exception thrown
	 * @param filename the name of the XML Instance Document that failed validation 
	 */
	private static void schemaValidationFail(SAXException e, String filename) {
		if (e instanceof SAXParseException) {
			SAXParseException saxException = (SAXParseException) e;
			System.err.println("\tResult: " + filename + " is not schema-valid.");
			System.err.println("\tError:  " + "Line: " + saxException.getLineNumber() + ", " + saxException.getMessage());
			//System.err.printf("Stack Trace:");
			//e.printStackTrace(System.err);
		} else {
			System.err.println(e.getMessage());
			System.err.printf("Stack Trace:");
			e.printStackTrace(System.err);
		}
	}

	/**
	 * This method uses the given command line arguments to 
	 * construct and configure a com.siemens.ct.exi.EXIFactory 
	 * object which will be used to either encode or decode EXI.
	 * 
	 * @param commandLineArgs a map of ExiOption enum values and String values (not all ExiOption enums require a String value)
	 */
	private static com.siemens.ct.exi.EXIFactory createEXIFactory(Map<ExiOption, String> commandLineArgs) {
		System.out.println("\tUsing the following EXI Options:");

		EXIFactory exiFactory = DefaultEXIFactory.newInstance(); // construct a basic com.siemens.ct.exi.EXIFactory
		try {

			FidelityOptions fidelityOptions = FidelityOptions.createDefault();

			// STRICT
			if (commandLineArgs.containsKey(ExiOption.STRICT)) {
				printEnabledMessage(ExiOption.STRICT);
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_STRICT, true);
			}
			else
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_STRICT, false);

			// PRESERVE_NAMESPACE_PREFIXES
			if (commandLineArgs.containsKey(ExiOption.PRESERVE_NAMESPACE_PREFIXES)) {
				printEnabledMessage(ExiOption.PRESERVE_NAMESPACE_PREFIXES);
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_PREFIX, true);
			}
			else
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_PREFIX, false);

			// PRESERVE_DTD
			if (commandLineArgs.containsKey(ExiOption.PRESERVE_DTD)) {
				printEnabledMessage(ExiOption.PRESERVE_DTD);
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_DTD, true);
			}
			else
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_DTD, false);

			// PRESERVE_LEXICAL_VALUES
			if (commandLineArgs.containsKey(ExiOption.PRESERVE_LEXICAL_VALUES)) {
				printEnabledMessage(ExiOption.PRESERVE_LEXICAL_VALUES);
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_LEXICAL_VALUE, true);
			}
			else
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_LEXICAL_VALUE, false);

			// PRESERVE_COMMENTS
			if (commandLineArgs.containsKey(ExiOption.PRESERVE_COMMENTS)) {
				printEnabledMessage(ExiOption.PRESERVE_COMMENTS);
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_COMMENT, true);
			}
			else
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_COMMENT, false);

			// PRESERVE_PIS
			if (commandLineArgs.containsKey(ExiOption.PRESERVE_PI)) {
				printEnabledMessage(ExiOption.PRESERVE_PI);
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_PI, true);
			}
			else
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_PI, false);

			// SELF_CONTAINED
			if (commandLineArgs.containsKey(ExiOption.SELF_CONTAINED)) {
				printEnabledMessage(ExiOption.SELF_CONTAINED);
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_SC, true);
			}
			else
				fidelityOptions.setFidelity(FidelityOptions.FEATURE_SC, false);

			exiFactory.setFidelityOptions(fidelityOptions); // tell the EXIFactory to use these FidelityOptions

			// CodingMode.COMPRESSION
			// CodingMode.BIT_PACKED is recommended for small messages, CodingMode.COMPRESSION is recommended for large messages
			if (commandLineArgs.containsKey(ExiOption.COMPRESSION)) {
				printEnabledMessage(ExiOption.COMPRESSION);
				exiFactory.setCodingMode(CodingMode.COMPRESSION);
			}
			else {
				// handle ALIGNMENT only when COMPRESSION is not specified
				if (commandLineArgs.containsKey(ExiOption.ALIGNMENT)) {
					CodingMode codingMode = null;
					try {
						codingMode = CodingMode.valueOf(commandLineArgs.get(ExiOption.ALIGNMENT).toUpperCase());
					} catch (Exception e) {
						System.out.println("\nError: the value '" + commandLineArgs.get(ExiOption.ALIGNMENT) + "' is not recognized. Permissible values are " + ExiOption.ALIGNMENT.getValuePlacemark());
						printHelp(); // this calls System.exit(0)
					}
					System.out.println("\t\t" + ExiOption.ALIGNMENT.getName() + " set to " + codingMode);
					exiFactory.setCodingMode(codingMode);
				}
			}

			// FRAGMENT
			if (commandLineArgs.containsKey(ExiOption.FRAGMENT)) {
				printEnabledMessage(ExiOption.FRAGMENT);
				exiFactory.setFragment(true);
			}

			// SCHEMA OPTIONS
			GrammarFactory grammarFactory = GrammarFactory.newInstance();
			Grammar grammar = null;
			if (commandLineArgs.containsKey(ExiOption.DECODING_SCHEMA)) {
				System.out.println("\t\t" + ExiOption.DECODING_SCHEMA.getName() + "...\n\t\t\tloading " + commandLineArgs.get(ExiOption.DECODING_SCHEMA));
				grammar = grammarFactory.createGrammar(commandLineArgs.get(ExiOption.DECODING_SCHEMA));
			}
			else if (commandLineArgs.containsKey(ExiOption.XSD_TYPES_ONLY)) {
				printEnabledMessage(ExiOption.XSD_TYPES_ONLY);
				grammar = grammarFactory.createXSDTypesOnlyGrammar();
			}
			else {
				System.out.println("\t\tSchemaless encoding/decoding will be used.");
				grammar = grammarFactory.createSchemaLessGrammar();
			}

			exiFactory.setGrammar(grammar);

			// BLOCK_SIZE
			if (commandLineArgs.containsKey(ExiOption.BLOCK_SIZE)) {
				int blockSize = Integer.parseInt(commandLineArgs.get(ExiOption.BLOCK_SIZE)); 
				System.out.println("\t\t" + ExiOption.BLOCK_SIZE.getName() + " = " + blockSize + " bytes");
				exiFactory.setBlockSize(blockSize);
			}

			// VALUE_MAX_LENGTH
			if (commandLineArgs.containsKey(ExiOption.VALUE_MAX_LENGTH)) {
				int valueMaxLength = Integer.parseInt(commandLineArgs.get(ExiOption.VALUE_MAX_LENGTH)); 
				System.out.println("\t\t" + ExiOption.VALUE_MAX_LENGTH.getName() + " = " + valueMaxLength + " bytes");
				exiFactory.setValueMaxLength(valueMaxLength);
			}

			// VALUE_PARTITION_CAPACITY
			if (commandLineArgs.containsKey(ExiOption.VALUE_PARTITION_CAPACITY)) {
				int valuePartitionCapacity = Integer.parseInt(commandLineArgs.get(ExiOption.VALUE_PARTITION_CAPACITY)); 
				System.out.println("\t\t" + ExiOption.VALUE_PARTITION_CAPACITY.getName() + " = " + valuePartitionCapacity + " bytes");
				exiFactory.setValuePartitionCapacity(valuePartitionCapacity);
			}

			System.out.println();

		} catch (UnsupportedOption e) {
			System.out.println("ERROR: UnsupportedOption: " + e.getMessage());
			System.exit(1);
		} catch (EXIException e) {
			System.out.println("ERROR: EXIException: " + e.getMessage());
			System.exit(1);
		}
		return exiFactory;
	}

	/**
	 * A small helper method to print an 'enabled' message to stdout.
	 * 
	 * @param exiOption the ExiOption to print
	 */
	private static void printEnabledMessage(ExiOption exiOption) {
		System.out.println("\t\t" + exiOption.getName() + " enabled");
	}

	/**
	 * Parses the command line arguments for ExiOptions and values.
	 * 
	 * @return a map of ExiOption enums and String values
	 */
	private static Map<ExiOption, String> parseCommandLineArgs(String[] args) {
		Map<ExiOption, String> commandLineArgs = new java.util.EnumMap<ExiOption, String>(ExiOption.class);

		int i = 0;
		while (i < args.length) {
			ExiOption exiOption = ExiOption.valueOfCommandLineArg(args[i].replaceFirst("-", ""));
			if (exiOption != null) {
				if (i + 1 < args.length && !exiOption.getValuePlacemark().isEmpty()) {
					commandLineArgs.put(exiOption, args[i + 1]);
					i = i + 2;
				}
				else if (!(i + 1 < args.length) && !exiOption.getValuePlacemark().isEmpty()) {
					System.out.println("\nError: option " + args[i] + " requires a value");
					printHelp(); // this calls System.exit(0)
				}
				else {
					commandLineArgs.put(exiOption, null);
					i++;
				}
			}
			else {
				System.out.println();
				System.out.println("ERROR: command line option '" + args[i] + "' not known");
				printHelp(); // this calls System.exit(0)
			}
		}

		System.out.println();
		return commandLineArgs;
	}

	/**
	 * Prints command-line help to stdout. This method calls System.exit(0).
	 */
	private static void printHelp() {
		String commandLineName = "java -jar ExiValidator.jar";
		System.out.println();
		System.out.println("Efficient XML Interchange (EXI) Validator");
		System.out.println("This program can be used to validate a text or EXI-encoded XML Instance Document against a XML Schema.");
		System.out.println();
		//System.out.println(tab + commandLineName + " [EXI instance document] [XML Schema file] [EXI Options to be used for decoding the EXI instance document]");
		System.out.println("Validate XML document against a schema:");
		System.out.println("\t" + commandLineName + " -" + ExiOption.XML_IN.getCommandLineArg() + " sample.xml -" + ExiOption.VALIDATION_SCHEMA.getCommandLineArg() + " sample.xsd");
		System.out.println();
		System.out.println("Validate EXI document against a schema:");
		System.out.println("\t" + commandLineName + " -" + ExiOption.EXI_IN.getCommandLineArg() + " sample.exi -" + ExiOption.VALIDATION_SCHEMA.getCommandLineArg() + " sample.xsd" + " [list of EXI Options to be used for decoding the EXI instance document]");
		System.out.println();
		System.out.println("EXI Options:");
		System.out.println();
		for (ExiOption exiOption : ExiOption.values()) {
			System.out.println("\t" + "-" + exiOption.getCommandLineArg() + " " + exiOption.getValuePlacemark());
			System.out.println("\t" + "\t" + exiOption.getDescription());
			if (exiOption.getUrl().length() > 0)
				System.out.println("\t" + "\t" + "URL: " + exiOption.getUrl());
			if (exiOption.getValuePlacemark().length() > 0)
				System.out.println("\t" + "\t" + "Example: " + exiOption.getExample());
			System.out.println();
		}
		System.exit(0);
	}

}
